<?php
/*
*
* class-wic-db-setup.php
*		accesses sql to create or update database and dictionary  
*/



class WIC_DB_Setup {

	/*********************************************************************************************
	*	
	*	database_setup()	
	*
	*	runs db delta to install/upgrade database
	*
	*  then runs routines to install/upgrade dictionary and options tables
	*
	*	note: all wic_issues_crm tables use the utf8 character set and utf8_general_ci collation  
	*
	*	note: version globals set in wp-issues-crm.php
	*
	*********************************************************************************************/
	private static function database_setup() { 

		global $wp_issues_crm_db_version; // see wp-issues-crm.php 
		$installed_version = get_option ( 'wp_issues_crm_db_version' ); 		
		
		global $wpdb;		

		// drop unnecessary index in version 3.3 -- do this before dbdelta so field lengths will take
		// but don't know what version upgrading from or if new, so check for tables and index before action
		if ( $installed_version < '3.3' ) { 
			$inbox_table = $wpdb->prefix . 'wic_inbox_image';
			$tables = $wpdb->get_results( "SHOW TABLES LIKE '$inbox_table'");
			if ( $tables ) {
				$indexes = $wpdb->get_results( "SHOW INDEX FROM $inbox_table");
				if ( $indexes ) {
					$do_delete = false;
					foreach ( $indexes as $index ) {
						if ( 'inbox_image' == $index->Key_name ) {
							$do_delete = true; 
							break;
						}
					}	
					if ( $do_delete ) {
						$wpdb->query ( "ALTER TABLE $inbox_table DROP INDEX inbox_image" );
					} // if do_delete
				} // if $indexes
			} // if $tables
		} // if < 3.3
		
		// load the sql for table set up  
		$sql = file_get_contents( plugin_dir_path ( __FILE__ ) . '../../sql/wic_structures.sql' );

		// add site specific db prefixing
		$sql = self::site_table_names_in_sql ( $sql );

		// do the table creation/upgrades
		require_once( ABSPATH . 'wp-admin/includes/upgrade.php' );
		$result = dbDelta( $sql, true ); // execute = true

		// if not present add a full text index to post_title on wp_post table
		$table = $wpdb->posts;		
		$sql = "SHOW INDEXES IN $table";
		$result = $wpdb->get_results ( $sql );
		$ok_index = false;
		foreach ( $result as $index_component ) {
			if ( 	'post_title' 	== $index_component->Column_name &&
					'FULLTEXT' 		== $index_component->Index_type &&
					1 				== $index_component->Seq_in_index	
				) {
					$ok_index = true;
					break;				
				}		
		}
		if ( ! $ok_index )  {
			$sql = "CREATE FULLTEXT INDEX wp_issues_crm_post_title ON $table ( post_title )";		
			$wpdb->query ( $sql );
		}
		
		/*
		* install or upgrade data dictionary and field groups
		* wic_data_dictionary_and_field_groups.sql executes the following steps:
		*   - lock dictionary and option tables
		* 	- first purge dictionary all except custom fields
		* 	- second truncate field groups		
		* 	- insert dictionary records 
		*	- insert field group records
		*	- unlock tables
		*/
		self::execute_file_sql ( 'wic_data_dictionary_and_field_groups' );

		// populate interface table if not populated, otherwise don't touch it
		$interface = $wpdb->prefix . 'wic_interface'; 
		$sql = "SELECT upload_field_name from $interface LIMIT 0, 1";
		$results = $wpdb->get_results ( $sql );
		if ( 0 == count ( $results ) ) {
			self::execute_file_sql ( 'wic_interface_table' );
		}

		// install base version of option groups if first install ( updated through 2.4 )
		if ( false === $installed_version ) {
			self::execute_file_sql ( 'wic_option_groups_and_options' );	
		}

		// add ons in version 2.5
		if ( false === $installed_version || $installed_version < '2.5' ) {
			self::execute_file_sql ( 'wic_option_groups_and_options_upgrade_005' );					
		}


		// add ons in version 2.6
		if ( false === $installed_version || $installed_version < '2.6' ) {
			self::execute_file_sql ( 'wic_option_groups_and_options_upgrade_006' );					
		}	

		// add ons in version 2.7
		if ( false === $installed_version || $installed_version < '2.7' ) {
			self::execute_file_sql ( 'wic_option_groups_and_options_upgrade_007' );					
		}
		
		// add ons in version 2.8
		if ( false === $installed_version || $installed_version < '2.8' ) {
			self::execute_file_sql ( 'wic_option_groups_and_options_upgrade_008' );					
		}
		// add ons in version 2.9
		if ( false === $installed_version || $installed_version < '2.9' ) {
			self::execute_file_sql ( 'wic_option_groups_and_options_upgrade_009' );					
		}
		// add ons in version 3.1
		if ( false === $installed_version || $installed_version < '3.1' ) {
			self::execute_file_sql ( 'wic_option_groups_and_options_upgrade_010' );					
		}

		// select becomes selectmenu in version 3.3
		if ( false === $installed_version || $installed_version < '3.3' ) {
			self::execute_file_sql ( 'wic_option_groups_and_options_upgrade_011' );					
		}

		// standardize autorecorded email type
		if ( false === $installed_version || $installed_version < '3.3.1' ) {
			self::execute_file_sql ( 'wic_option_groups_and_options_upgrade_012' );					
		}

		// change option set for issue display drop down
		if ( false === $installed_version || $installed_version < '3.4' ) {
			self::execute_file_sql ( 'wic_option_groups_and_options_upgrade_013' );					
		}
		
		// change option set for issue display drop down
		if ( false === $installed_version || $installed_version < '3.4.0.5000' ) {
			self::execute_file_sql ( 'wic_option_groups_and_options_upgrade_014' );					
		}
		/*
		* back fill option_group_id based on parent_option_group_slug for upgrades to option table 
		*  -- note that this query will include any groups user added between upgrades, since parent_option_group_slug not maintained elsewhere 
		*		
		* parent_option_group_slug is used only in the upgrade process (and one hard_coded reference in class-wic-db-access-dictionary.php)
		*
		* this step must run after all upgrades/additions to option_value and option_group
		*
		* it maintains integrity of multivalue logic (which relies on ID) while allowing upgrade 
		* to make additions to both option_group table and option_value table with indeterminate option_group_id's
		*
		* insert statements to option_value in upgrades should include parent_option_group_slug, but not option_group_id
		*/
		$option_group = $wpdb->prefix . 'wic_option_group'; 
		$option_value = $wpdb->prefix . 'wic_option_value';
		$sql = "UPDATE $option_value v INNER JOIN $option_group g ON v.parent_option_group_slug = g.option_group_slug 
			SET v.option_group_id = g.ID where option_group_id = '' ";
		$wpdb->query ( $sql );

		// force a run of role set up attached to activation hook (to reflect role definition change)
		if (  $installed_version < '2.4.2' ) {
			WIC_Admin_Setup::wic_set_up_roles_and_capabilities();			
		}

		// always finish by marking version change
		if ( false !== $installed_version ) {
			update_option( 'wp_issues_crm_db_version', $wp_issues_crm_db_version );		
		} else {
			add_option( 'wp_issues_crm_db_version', $wp_issues_crm_db_version );
		}
		
	}
	
	public static function update_db_check () { 

		global $wp_issues_crm_db_version; // see wp-issues-crm.php

		// check if database up to date; if not, run setup 
		// single version check covers database and dictionary -- unlikely that increase churn by combining
		$installed_version = get_option( 'wp_issues_crm_db_version');
		if ( $wp_issues_crm_db_version != $installed_version ) { // returns false if absent, so also triggered on first run
			self::database_setup();
		}

	}	
	
	public static function execute_file_sql ( $file_name ) {		
		global $wpdb;		
		
		// load the table set up sql 
		$sql = file_get_contents( plugin_dir_path ( __FILE__ ) . '../../sql/' . $file_name . '.sql' );
		// add local file prefixes
		$sql = self::site_table_names_in_sql ( $sql );
		
		// execute statements one by one 
		$sql_array = explode ( ';', $sql );
		$outcome = true;
		foreach ( $sql_array as $sql_statement ) {
			if ( $sql_statement > '' ) {
				$result = $wpdb->query ( $sql_statement );
				if ( false === $result ) {
					$outcome = false;			
				}
			}
		}
		return ( $outcome );
		
	}
	
	// replace standard prefix with possible site table prefix
	private static function site_table_names_in_sql ( $sql ) {
		global $wpdb;
		$sql = str_replace ( ' wp_wic_', ' ' . $wpdb->prefix . 'wic_' , $sql );		
		return ( $sql );
	}
}

