<?php
/*
*
*	wic-entity-email-deliver.php
*
*/
class WIC_Entity_Email_Deliver {
	
	/*
	*	Notes re queuing model for email sends -- replies and new sends
	*
	*	The act of replying to a message (or sending new) creates a new activity record with activity_type = wic_reserved_99999998.    
	*
	*	The queue of outgoing messages is just the set of activities with reserved wic_reserved_99999998.  If they are processed successfully,
	*		they get a new type, wic_reserved_99999999.  They are also stamped with an attempted time so that the same run will
	*		not see them -- selects attempted time not equal to own stamp ( which is fixed at the start of the run ).
	*
	*	Have to carefully control to prevent two queue processors running simultaneously -- see notes further below.
	*
	*
	*/
	private static function set_up_mailer() {
        // https://github.com/PHPMailer/PHPMailer/blob/master/examples/mailing_list.phps
        // https://github.com/PHPMailer/PHPMailer/wiki/Troubleshooting
		try { // try catch example https://github.com/PHPMailer/PHPMailer/blob/master/examples/exceptions.phps
			ob_start();

			// use wordpress set zone unless not set to a city
			date_default_timezone_set( get_option('timezone_string') > '' ? get_option('timezone_string') : 'Etc/UTC' );

			// initialize up mailer (get our own instance -- don't interfere with other WP use)
			require_once ABSPATH . WPINC . '/class-phpmailer.php';
			require_once ABSPATH . WPINC . '/class-smtp.php';
			$mail = new PHPMailer( true );
		
			// get settings
			$wic_settings = get_option( 'wp_issues_crm_plugin_options_array' );
		
			// apply settings relevant to mail send tool
			switch ( $wic_settings['email_send_tool'] ) {
				case 'smtp':
					$mail->isSMTP();
					$mail->Host 		= $wic_settings['email_smtp_server']; // mail.example.com
					$mail->Port 		= $wic_settings['smtp_port']; 		// 587 or 25 with tls or 465 for ssl
					$mail->SMTPSecure 	= $wic_settings['smtp_secure']; 		// tls or ssl or not; (use tls)
					$mail->SMTPAutoTLS 	= false; // always -- don't jump into TLS mode by accident
					$mail->SMTPAuth 	= true; 							// always
					$mail->Username 	= $wic_settings['smtp_user']; 		// mail_user@example.com;
					$mail->Password 	= $wic_settings['smtp_password']; 	// 
					if ( 'no' == $wic_settings['require_good_ssl_certificate'] ) {
						$mail->SMTPOptions = array(							// accept unsigned certificates
							'ssl' => array(
								'verify_peer' => false,
								'verify_peer_name' => false,
								'allow_self_signed' => true
							)
						);		
					} elseif ('selfOK' == $wic_settings['require_good_ssl_certificate'] ){
						$mail->SMTPOptions = array(
							'ssl' => array(
								'allow_self_signed' => true,
								'peer_name' => ( $wic_settings['peer_name'] > '' ? $wic_settings['peer_name'] : $wic_settings['email_smtp_server'] ) ,
							)
						);
					} // if not either of the lower level security options, go with defaults 

					if ( 'yes' == $wic_settings['use_IPV4'] ) {
						$mail->Host = gethostbyname ( $wic_settings['email_smtp_server'] );
					}

					break;
				case 'mail':
					$mail->IsMail();
					break;
				case 'sendmail':
					$mail->isSendmail();
					break;
			}
		
			// settings applicable to all tools
			$mail->SMTPDebug 	= $wic_settings['smtp_debug_level']; 	
			$mail->Debugoutput = 'html';
			$mail->CharSet = 'utf-8'; // this is wishful thinking -- after layers of phpmailer and smtp server, this can come out as ascii or possibly iso-8859								
			$mail->setFrom 	  ( $wic_settings['from_email'] ? $wic_settings['from_email'] : $wic_settings['smtp_user'], $wic_settings['from_name'] );
			$mail->addReplyTo(  $wic_settings['smtp_reply'], $wic_settings['reply_name']);
			
			ob_end_clean();
			return ( $mail );

		}  catch (phpmailerException $e) {
			ob_end_clean();
    		return ( $e->errorMessage() ); 	// error messages from PHPMailer
		} catch (Exception $e) {
			ob_end_clean();
    		return ( $e->getMessage() ); 	// error messages from anything else!
		}

	}
	
	
	// send test email from settings screen
	public static function test_settings ( $dummy, $email_address ) {
		
		// get a mailer
		$mail = self::set_up_mailer();
		
		$output			= 'success';
		if ( ! is_object( $mail ) ) { 
			return array ( 'response_code' => true, 'output' => '<br/><strong>Error in mailer setup.</strong><br/>' . $mail );
		} else {
			// set up output buffer to capture debug dialog 
			ob_start();
			try {
				// set up generic test settings
				$mail->Subject = 'Test Email from WP Issues CRM';
				$body = 
					'<h3>Great job!</h3><p>Your settings are good.</p><br/>' . 
					'<p>You have successfully sent a test email from WP Issues CRM.</p>' .
					'<p>Reply to this message to test your reply email setting.</p>';
				$mail->msgHTML($body);	
				$mail->addAddress( $email_address, $email_address ); 	

				// send mail	 
				$mail->send();  
			} catch ( phpmailerException $e ) { 
				$output = '<br/><strong>Not successful. </strong><br/><br/>'. $e->errorMessage() .  ob_get_contents() ; 	
			} catch (Exception $e) {
				$output = '<br/>Not successful -- general error. <br/><br/>'. $e->getMessage() .  ob_get_contents() ; 	
			}
			ob_end_clean();
		}
		return array ( 'response_code' => true, 'output' => $output );
	}


	/*
	*
	*  Mailer control functions
	*/	
	public static function get_mailer_status( $dummy1, $dummy2 ) {
		return array ( 'response_code' => true, 'output' => get_option ( 'wp-issues-crm-mailer-ok-to-go', true ) );  // start with mailer released if have not set option
	} 

	public static function set_mailer_status( $dummy, $new_status ) {
		update_option ( 'wp-issues-crm-mailer-ok-to-go', $new_status );
		return array ( 'response_code' => true, 'output' =>  $new_status );
	} 

	// initiated only as scheduled cron -- see wp-issues-crm.php
	public static function process_message_queue () {
		if ( WP_DEBUG ) {
			error_log ( 'WP Issues CRM cron run: Starting process_message_queue.' );
		}
		// get processing options
		$form_variables_object = unserialize ( get_option ( 'wp-issues-crm-email-processing-options' ) );
		$wic_settings = get_option( 'wp_issues_crm_plugin_options_array' );
		$folder = $wic_settings['imap_inbox'];
		
		/*
		*	The mailer is protected against double submits.  
		*	Cron could double submit runs if delayed.
		*
		*	It is designed to run based on an hourly schedule.
		*
		*	Protections against double submits are in two layers
		*		+ Timestamp will not allow start if another has not terminated normally 
		*			( will allow after 4000 seconds elapsed since last start )
		*		+ If that fails ( e.g., prior run is actually continuing over 4000 seconds )
		*			- a message that appears to have been double sent will immediately halt the job.
		*			- a message that appears to have failed twice will immediately halt the job
		*	
		*	The timestamping for the send process does not happen until the process_mails completes, so, in effect, it is a separate process, possibly thrown off schedule.
		*	Since start scheduling (but not double-run protections) coupled, can push mailer start back by long process_mails.  If that happens, 
		*	   mailer still will not start unless last run (however off schedule) is more than 4000 seconds ago (i.e., likely terminated );
		*
		*/
		// set time_stamp
		$time_stamp = time();
		
		// check for possibility that another run is ongoing
		$last_stamp = get_option ( 'wp_issues_crm_send_mail_running_timestamp' );
		/*
		* If last stamp is over one hour ago ( or negative from normal termination ),
		*	assume that not running;
		*
		*/
		if ( $last_stamp < $time_stamp - 4000 ) {
			update_option ( 'wp_issues_crm_send_mail_running_timestamp', $time_stamp );
		} else {
			if ( WP_DEBUG ) {
				error_log ( 'Terminating WP Issues Cron Run because previous run not terminated or terminated abnormally; will resolve after max wait time of 4000 seconds.' );
			}
			exit (0);
		}
		
		// get delay time setting
		$delay = $wic_settings['send_mail_sleep_time'];
		
		// set up mailer with all constant parameters
		$mail = self::set_up_mailer();
		if ( !is_object ( $mail ) ) {
			update_option ( 'wp_issues_crm_send_mail_running_timestamp', -1 );
			if ( WP_DEBUG ) {
				error_log ( 'No mailer available to WP Issues Cron Run -- terminating run.  Check settings.' );
			}
			exit (0);
		}
		// use persistent setting only if option selected (close at end based on option too)
		$wic_settings = get_option( 'wp_issues_crm_plugin_options_array' );
		$persistent = isset ( $wic_settings['use_persistent_connection'] ) ? ( 1 == $wic_settings['use_persistent_connection'] ) : false;
		$mail->SMTPKeepAlive = $persistent;

		$mail->SMTPDebug = 0;
		/*
		* loop while there are records in the queue
		*
		* do not simply copy the queue  to an array and loop through it -- that would not allow things to be added to the queue 
		*	also would result in higher risks if double process somehow occurred
		*
		* grab a record ( activity_type is queued message, not already timestamped as attempted by this run, oldest )
		* timestamp it as attempted ( done by get_next_queued_message )
		*
		*/
		$success_count = 0;
		$failure_table = '';
		$normal_termination = true;
		$failed_messages = array(); // used to test if overlap has occurred
		while ( $message = self::get_next_queued_message( $time_stamp ) ) {
		
			// test to see if OK to continue running
			if ( ! get_option ( 'wp-issues-crm-mailer-ok-to-go', true ) ) {
				$normal_termination = false;
				break;
			}
			
			// give ourselves some time -- continually extend run time limit as long as actually processing records 
			// allow more than enough time for SMTP delays
			set_time_limit( 60 + $delay * 1000 );		
			
			// attempt to send it
			try {
				// set up output buffer to capture debug dialog 
				ob_start();
				/*
				* set addresses
				* note that mailer will bounce an unset to address and will capture below
				*/
				foreach ( $message->to_array as $address ) {
					$mail->addAddress( $address[1], $address[0] ); // adds TO address
				}
				foreach ( $message->cc_array as $address ) {
					$mail->addCC( $address[1], $address[0] );
				}
				foreach ( $message->bcc_array as $address ) {
					$mail->addBCC( $address[1], $address[0] );
				}
				/*
				*
				* subject and body
				*
				*/				
				$mail->Subject = $message->subject;
				$mail->msgHTML( $message->html_body );
				$mail->AltBody = $message->text_body;  // For short notes, this is better than phpMailer html to text conversion	
				/*
				*
				* add any attachments -- note, including them in reply as well as cc
				*
				* note that if inbox has been purged, will not find attachments; no risk of wrong attachments -- links deleted on purge
				*/
				if ( $attachments = self::check_for_attachments ( $message->is_reply_to ) ) {
					foreach ( $attachments as $attachment ) {
						$mail->addStringAttachment( 
							$attachment->attachment, // BLOB content
							$attachment->attachment_filename, // as parsed from image
							'base64', // default transfer encoding
							$attachment->attachment_type . ( $attachment->attachment_subtype ?  '/' . $attachment->attachment_subtype : '' ), // type/subtype
							'attachment' // disposition
						);					
					}
				}			

	
				// send mail	 
				$mail->send();  
				$response = 'success';
			} catch ( phpmailerException $e ) { 
				$response = '<br/><strong>Not successful. </strong><br/><br/>'. $e->errorMessage() .  ob_get_contents() ; 	
			} catch (Exception $e) {
				$response = '<br/>Not successful -- general error. <br/><br/>'. $e->getMessage() .  ob_get_contents() ; 	
			}
			ob_end_clean();
			
			// if successful, take it off the queue by changing activity_type
			if ( 'success' == $response ) {
				$success_count++;
				if ( ! self::unqueue_message ( $message->ID ) ) {
					if ( WP_DEBUG ) {
						error_log ( 'Terminating WP Issues Cron Run because another copy running.  One message probably double sent.' );
					}
					exit(0);
				}				
			} else {
				// test to see if message has alread failed once -- if so quit, because shouldn't be seeing this unless another 
				// copy is changing the send attempt timestamp on the record
				if ( in_array( $message->ID, $failed_messages ) ) {
					if ( WP_DEBUG ) {
						error_log ( 'Terminating WP Issues Cron Run because another copy running.  No messages were double sent' );
					}
					exit(0);				
				} else {
					array_push ( $failed_messages, $message->ID );
				}
				$failure_table .=
				'<tr>' .
					'<td>' . $message->ID . '</td>' .
					'<td>' . $message->to_array[0][1] . '</td>' .
					'<td>' . $message->subject . '</td>' .
					'<td>' . $response . '</td>' .
				'</tr>'; 
			}
			
			// clear addresses, https://phpmailer.github.io/PHPMailer/classes/PHPMailer.html#method_clearAddresses 
			$mail->clearBCCs(); // bcc addresses
			$mail->clearCCs(); // cc addresses
			$mail->clearAddresses(); // To Addressses
			$mail->clearAttachments(); 
			
			// delay at end of loop for pacing (elsewhere in loop, greater probability of a double send occurring in overlapping runs )
			usleep ( $delay * 1000 ); 
		}				
		
		// update run flag to reflect termination
		update_option ( 'wp_issues_crm_send_mail_running_timestamp', -1 );

		// send report to administrators
		self::send_mailing_report ( $mail, $time_stamp, $success_count, $failure_table, $normal_termination );
		if ( WP_DEBUG ) {
			error_log ( 'WP Issues CRM cron run: Completed process_message_queue.' );
		}
		// close connection if was using open connection through loop
		if ( $persistent ) { 
			$mail->smtpClose();
		}; 
		
		
	} // close process_message_queue()

	private static function send_mailing_report ( &$mail, $time_stamp, $success_count, $failure_table, $termination_normal ) {

		// after all records sent, report completion of run	
		try {
		
			// get report address -- if not set, no report
			$wic_settings = get_option( 'wp_issues_crm_plugin_options_array' );
			if ( $wic_settings['send_mail_reports'] ) {
				$mail->addAddress( $wic_settings['send_mail_reports'], $wic_settings['send_mail_reports'] );
			} else {
				return('');
			}

			$mail->Subject = 'WP Issues CRM Mailing Report -- Mailing Run: ' . date( 'Y-m-d H:i:s', $time_stamp );
			$outcome_message = $termination_normal ?
				'The mailing run submitted on the following date and time ran to normal completion: ' :
				'The mailing run terminated early because it was suspended.';
			$body = '<h3>Mailing Report from site <a href="' . site_url() . '">' . get_bloginfo() . '</a></h3>' .
					'<p>'. $outcome_message .  date( DATE_RFC2822, $time_stamp ) . '.</p>' .
					'<p>The number of messages successfuly sent was: '. $success_count . '.</p>';
			if ( $failure_table > '' ) {
				$body .= '<p>The following messages were processed but not sent successfully.</p>' .
					'<table>
						<colgroup>
							<col style="width:5%">
							<col style="width:25%">
							<col style="width:40%">
							<col style="width:30%">
							
					 	</colgroup>' .  
					'<tbody> .
						<tr>
							<th>ID</th>
							<th>To</th>
							<th>Subject</th>
							<th>Errors</th>
						</tr>' . 
						$failure_table . 
					'</tbody></table>';
			} else {
				$body .= '<p>There were no messages that failed to send.</p>';
			}
		
			$body .= '<p><a href="' . site_url() . '/wp-admin/admin.php?page=wp-issues-crm-email' .'">View your email batch log here.</a></p>';
			$mail->msgHTML ( $body );
			$mail->send();
			// can't do much here; if user isn't seeing messages should go to test set up
		} catch ( phpmailerException $e ) { 
		} catch (Exception $e) {
		}

	}
	
	// retrieve next queued message and stamp it as attempted.
	private static function get_next_queued_message( $time_stamp ) {
		global $wpdb;
		$outbox_table = $wpdb->prefix . "wic_outbox";
		// get next queued
		$sql_first = "
			SELECT ID, serialized_email_object from $outbox_table  
			WHERE sent_ok = 0 AND held = 0
			AND attempted_send_time_stamp != $time_stamp 
			ORDER BY queued_date_time
			LIMIT 0, 1
			";
		$results = $wpdb->get_results ( $sql_first );
		if ( isset ( $results[0] ) ) {
			$next_ID = $results[0]->ID;
		} else {
			return ( false );
		}
		// stamp next queued
		$sql_stamp = "UPDATE $outbox_table SET attempted_send_time_stamp = $time_stamp WHERE ID = $next_ID"; 
		$wpdb->query ( $sql_stamp );

		// return next queued message object
		$message = unserialize ( $results[0]->serialized_email_object );
		$message->ID = $results[0]->ID;
		return $message;
	}

	private static function check_for_attachments ( $is_reply_to ) {
		if ( ! $is_reply_to ) {
			return false;
		}
		global $wpdb;
		$attachments_table = $wpdb->prefix . 'wic_inbox_image_attachments';	
		// note that, on upgrade, might lose a couple of attachments in first cycle if messages pending without full_folder_string	
		$result = $wpdb->get_results ( "SELECT * from $attachments_table WHERE message_id = $is_reply_to " );
		return $result;
	}
	
	// move message from send queue to sent record 
	private static function unqueue_message ( $id ) {
		global $wpdb;
		$outbox_table = $wpdb->prefix . "wic_outbox";
		// first test to see if message is already unqueued by another copy of the job and return false if so
		$sql_test = "SELECT ID FROM $outbox_table WHERE ID = $id AND sent_ok = 1";
		if ( $wpdb->get_results ( $sql_test ) ) {
			return false;
		}
		// otherwise, do the update and return true 
		$sql = "UPDATE $outbox_table SET sent_ok = 1, sent_date_time = NOW() WHERE ID = $id"; 
		$wpdb->query ( $sql );
		return true;
	}
	
}